import math
from typing import Callable, Any
from uuid import uuid4

import networkx as nx
import pandas as pd

from sampo.generator.pipeline.project import get_start_stage, get_finish_stage
from sampo.schemas.contractor import Contractor
from sampo.schemas.graph import GraphNode, WorkGraph, EdgeType
from sampo.schemas.requirements import WorkerReq
from sampo.schemas.resources import Worker
from sampo.schemas.works import WorkUnit

UNKNOWN_CONN_TYPE = 0
NONE_ELEM = '-1'


def break_circuits_in_input_work_info(works_info: pd.DataFrame) -> pd.DataFrame:
    """
    The function breaks circuits:
    - find all circuits,
    - break first edge in circle, for example,

    we have edges:
    (1-2), (2-3), (3-4), (4-2)
    function deletes edge (2-3)

    :param works_info: dataframe, that contains information about works
    :return: cleaned work_info
    """
    circuits = find_all_circuits(works_info)

    for _, row in enumerate(works_info.index[::-1]):
        for cycle in circuits:
            inter = list(set(row['predecessor_ids']) & set(cycle))
            if len(inter) != 0:
                index = row['predecessor_ids'].index(inter[0])
                row['predecessor_ids'].pop(index)
                row['connection_types'].pop(index)
                row['lags'].pop(index)

    return works_info


def find_all_circuits(works_info: pd.DataFrame) -> list[list[str]]:
    """
    The function find all elementary circuits using the algorithm of Donald B. Johnson
    doi: 10.1137/0205007

    :param works_info: dataframe, that contains information about works
    :return: list of cycles
    """
    graph = nx.DiGraph()
    edges = []

    for _, row in works_info[::-1].iterrows():
        v = row['activity_id']
        for u in row['predecessor_ids']:
            edges.append((v, u))

    graph.add_nodes_from(list(works_info['activity_id']))
    graph.add_edges_from(edges)

    return list(nx.simple_cycles(graph))


def fix_df_column_with_arrays(column: pd.Series, cast: Callable[[str], Any] | None = str,
                              none_elem: Any | None = NONE_ELEM) -> pd.Series:
    new_column = column.copy().astype(str).apply(
        lambda elems: [cast(elem) for elem in elems.split(',')] if elems != str(math.nan) else [none_elem])
    return new_column


def preprocess_graph_df(frame: pd.DataFrame) -> pd.DataFrame:
    def normalize_if_number(s):
        return str(int(float(s))) \
            if s.replace('.', '', 1).isdigit() \
            else s

    frame['activity_id'] = frame['activity_id'].astype(str)
    frame['volume'] = frame['volume'].astype(float)

    frame['predecessor_ids'] = fix_df_column_with_arrays(frame['predecessor_ids'], cast=normalize_if_number)
    frame['connection_types'] = fix_df_column_with_arrays(frame['connection_types'],
                                                          cast=EdgeType,
                                                          none_elem=EdgeType.FinishStart)
    if 'lags' not in frame.columns:
        frame['lags'] = [NONE_ELEM] * len(frame)
    frame['lags'] = fix_df_column_with_arrays(frame['lags'], float)

    frame = break_circuits_in_input_work_info(frame)

    return frame


def add_graph_info(frame: pd.DataFrame) -> pd.DataFrame:
    existed_ids = set(frame['activity_id'])

    def filter_predecessor_info(row):
        filtered_predecessors = [(pred_id, con_type, lag)
                                 for pred_id, con_type, lag in zip(row['predecessor_ids'], row['connection_types'], row['lags'])
                                 if pred_id in existed_ids]
        if not filtered_predecessors:
            filtered_predecessors.append((NONE_ELEM, EdgeType.FinishStart, float(NONE_ELEM)))
        return filtered_predecessors

    frame['edges'] = frame.apply(filter_predecessor_info, axis=1)
    frame.drop(['predecessor_ids', 'connection_types', 'lags'], axis=1, inplace=True)

    return frame


def topsort_graph_df(frame: pd.DataFrame) -> pd.DataFrame:
    """
    Sort DataFrame of works in topological order

    :param frame: DataFrame of works
    :return: topologically sorted DataFrame
    """
    G = nx.DiGraph()
    for _, row in frame.iterrows():
        G.add_node(row['activity_id'])
        for pred_id, con_type, lag in row['edges']:
            G.add_edge(pred_id, row['activity_id'])

    sorted_nodes = list(nx.topological_sort(G))
    frame['sort_key'] = frame['activity_id'].apply(lambda x: sorted_nodes.index(x))
    frame = frame.sort_values('sort_key')

    return frame


def build_work_graph(frame: pd.DataFrame, resource_names: list[str]) -> WorkGraph:
    start = get_start_stage()
    has_succ = set()
    id_to_node = {NONE_ELEM: start}

    for _, row in frame.iterrows():
        if 'min_req' in frame.columns and 'max_req' in frame.columns:
            reqs = [WorkerReq(res_name, row[res_name],
                              row['min_req'][res_name],
                              row['max_req'][res_name]
                              ) for res_name in resource_names
                    if 0 < row['min_req'][res_name] <= row['max_req'][res_name]]
        else:
            reqs = [WorkerReq(kind=res_name,
                              volume=row[res_name],
                              min_count=int(row[res_name] / 3),
                              max_count=math.ceil(row[res_name] * 10))
                    for res_name in resource_names
                    if row[res_name] > 0]
        is_service_unit = len(reqs) == 0
        work_unit = WorkUnit(row['activity_id'], row['activity_name'], reqs, group=row['activity_name'],
                             volume=row['volume'], volume_type=row['measurement'], is_service_unit=is_service_unit,
                             display_name=row['activity_name'])
        has_succ |= set(row['edges'][0])
        parents = [(id_to_node[p_id], lag, conn_type) for p_id, conn_type, lag in row.edges]
        node = GraphNode(work_unit, parents)
        id_to_node[row['activity_id']] = node

    without_succ = list(set(id_to_node.keys()) - has_succ)
    without_succ = [id_to_node[index] for index in without_succ]
    end = get_finish_stage(without_succ)
    graph = WorkGraph(start, end)
    return graph


def get_graph_contractors(path: str, contractor_name: str | None = 'ООО "***"') -> (
        list[Contractor], dict[str, float]):
    contractor_id = str(uuid4())
    workers_df = pd.read_csv(path, index_col='name')
    workers = {(name, 0): Worker(str(uuid4()), name, count * 2, contractor_id)
               for name, count in zip(workers_df.index, workers_df['count'])}
    workers_max_count = workers_df['count'].to_dict()
    contractors = [Contractor(contractor_id, contractor_name, list(workers_max_count.keys()), [], workers, {})]
    return contractors, workers_max_count



from itertools import chain
from uuid import uuid4

import pandas as pd
from pandas import DataFrame

from sampo.schemas.contractor import Contractor, get_contractor_for_resources_schedule
from sampo.schemas.graph import WorkGraph
from sampo.schemas.resources import Worker
from sampo.schemas.time_estimator import DefaultWorkEstimator, WorkTimeEstimator
from sampo.userinput.parser.exception import InputDataException
from sampo.userinput.parser.general_build import add_graph_info, topsort_graph_df, build_work_graph, preprocess_graph_df
from sampo.userinput.parser.history import set_connections_info
from sampo.utilities.task_name import NameMapper


class CSVParser:

    @staticmethod
    def work_graph_and_contractors(project_info: str,
                                   history_data: str | None = None,
                                   contractor_info: str | None = None,
                                   contractor_types: list[int] | None = None,
                                   unique_work_names_mapper: NameMapper | None = None,
                                   work_resource_estimator: WorkTimeEstimator = DefaultWorkEstimator(),
                                   contractors_number: int = 1) \
            -> (WorkGraph, Contractor):
        """
        Gets a WorkGraph and Contractors from file .csv.

        Schema of WorkGraph .csv file:
            mandatory fields:
                activity_id: str - Id of the current task,
                measurement: str - Measure of the size of the current task (e.g., km, pcs, lit),
                volume: float - Volume of the current task
            optional fields:
                granular_name: str - Task name as in the document,
                predecessor_ids: list[str] - Ids of predecessors of the current task,
                connection_types: list[str] - Types of links between the current task and its predecessors,
                lags: float - Time lags,
                min_req: dict[str: float] - A dictionary containing the minimum amount of each resource that is required to perform the current task
                max_req: dict[str: float] - A dictionary containing the maximum amount of each resource that is required to perform the current task

        Schema of Contractors .csv file (optional data):
            mandatory fields:
                contractor_id: str - Id of the current contractor,
                name: str - Contractor name as in the document
            optional fields:
                {names of resources}: float - each resource is a separate column

        Schema of history .csv file (optional data):
            mandatory fields:
                granular_smr_name: str - Task name as in the document,
                first_day: str - Date of commencement of the work,
                last_day: str - Date of completion
                upper_works: list[str] - Names of predecessors of the current task

        ATTENTION!
            1) If you send WorkGraph .csv file without data about connections between tasks, you need to provide .csv
            history file - the SAMPO will be able to reconstruct the connections between tasks based on historical data.
            2) If you do not provide work resource estimator, framework uses built-in estimator


        :param contractor_types:
        :param contractors_number: if we do not receive contractors, we need to know how many contractors the user wants,
        for a generation
        :param history_data: path to history data .csv file
        :param contractor_info: path to contractor info .csv file
        :param work_resource_estimator: work estimator that finds necessary resources, based on the history data
        :param project_info: path to .csv file
        :return: WorkGraph
        """

        if contractor_types is None:
            contractor_types = [25]

        graph_df = pd.read_csv(project_info, sep=';', header=0) if isinstance(project_info,
                                                                              str) else project_info
        # 1. work with tasks' connections. Preprocessing and breaking loops
        if history_data is None or (not(history_data is None) and 'predecessor_ids' in graph_df.columns):
            # 1.1. if we have info about predecessors in graph_df
            if 'predecessor_ids' not in graph_df.columns:
                raise InputDataException(
                    'you have neither history data about tasks nor tasks\' connection info in received .csv file.')
            works_info = preprocess_graph_df(graph_df)
        else:
            # 1.2. if we ought to restore predecessor info from history data
            history_df = pd.read_csv(history_data)
            graph_df = set_connections_info(graph_df, history_df)
            works_info = preprocess_graph_df(graph_df)

        # 2. gather resources and contractors based on work resource estimator or contractor info .csv file
        contractors = []
        resource_names = []
        works_info['activity_name_original'] = works_info.activity_name
        if unique_work_names_mapper:
            works_info.activity_name = works_info.activity_name.apply(lambda name: unique_work_names_mapper[name])

        if contractor_info is None:
            resources = [work_resource_estimator.find_work_resources(w[0], float(w[1]))
                         for w in works_info.loc[:, ['activity_name', 'volume']].to_numpy()]
            contractors = [get_contractor_for_resources_schedule(resources,
                                                                 contractor_capacity=contractor_types[i],
                                                                 contractor_id=str(i),
                                                                 contractor_name='Contractor' + ' ' + str(i + 1))
                           for i in range(contractors_number)]
        else:
            # if contractor info is given or contractor info and work resource estimator are received simultaneously
            contractor_df = pd.read_csv(contractor_info, sep=';', header=0) if isinstance(contractor_info,
                                                                                          str) else contractor_info
            for index, row in contractor_df.iterrows():
                specified_id = 'contractor_id' in contractor_df.columns
                column_lag = 2 if specified_id else 1
                contractor_id = row['contractor_id'] if specified_id else str(uuid4())
                contractors.append(
                    Contractor(id=contractor_id,
                               name=row['name'],
                               workers={item[0]: Worker(str(ind), str(item[0]), int(item[1])) for ind, item in
                                        enumerate(row[column_lag:].items())},
                               equipments=dict())
                )
            resource_names = contractor_df.columns[1:].to_list()
            if len(contractors) == 0 and isinstance(work_resource_estimator, DefaultWorkEstimator):
                raise InputDataException(
                    'you have neither info about contractors nor work resource estimator.'
                )
            resources = [work_resource_estimator.find_work_resources(w[0], float(w[1]), resource_names)
                         for w in works_info.loc[:, ['activity_name', 'volume']].to_numpy()]

        unique_res = list(set(chain(*[r.keys() for r in resources])))
        works_info.loc[:, unique_res] = DataFrame(resources).fillna(0)

        works_resources = add_graph_info(works_info)
        works_resources = topsort_graph_df(works_resources)
        work_graph = build_work_graph(works_resources, unique_res)

        # if we have no info about contractors or the user send an empty .csv file
        if len(contractors) == 0:
            contractors = [get_contractor_for_resources_schedule(resources,
                                                                 contractor_id=str(i),
                                                                 contractor_name='Contractor' + ' ' + str(i + 1))
                           for i in range(contractors_number)]

        return work_graph, contractors



from datetime import datetime
from typing import Tuple

import numpy as np
import pandas as pd

from sampo.utilities.task_name import NameMapper


def get_all_connections(tasks_df: pd.DataFrame, use_mapper: bool = False, mapper: NameMapper = None) \
        -> Tuple[dict[str, list], dict[str, list]]:
    """
    Return all connections from DataFrame of works
    """

    task_name_column = 'activity_name'
    if 'granular_name' in tasks_df:
        task_name_column = 'granular_name'

    num_tasks = len(tasks_df)
    # Get the upper triangular indices to avoid duplicate pairs
    indices = np.triu_indices(num_tasks, k=1)
    works1_ids = tasks_df['activity_id'].values[indices[0]]
    works1_names = tasks_df[task_name_column].values[indices[0]]

    works2_ids = tasks_df['activity_id'].values[indices[1]]
    works2_names = tasks_df[task_name_column].values[indices[1]]

    if use_mapper:
        works1_names = np.vectorize(mapper.get)(works1_names)
        works2_names = np.vectorize(mapper.get)(works2_names)

    return {"ids": works1_ids, "names": works1_names}, {"ids": works2_ids, "names": works2_names}


def get_delta_between_dates(first: str, second: str) -> int:
    return max((datetime.date(int(first.split('-')[0]), int(first.split('-')[1]), int(first.split('-')[2])) -
                datetime.date(int(second.split('-')[0]), int(second.split('-')[1]), int(second.split('-')[2]))).days, 1)


def find_min_without_outliers(lst: list[float]) -> float:
    return round(min([x for x in lst if x >= np.mean(lst) - 3 * np.std(lst)]), 2)


def gather_links_types_statistics(s1: str, f1: str, s2: str, f2: str) \
        -> Tuple[int, int, int, list, list, int, list, list, int, list, list, int, list, list]:
    """
    Count statistics on the occurrence of different mutual arrangement of tasks

    :param s1: start of first work
    :param f1: finish of first work
    :param s2: start of second work
    :param f2: finish of second work
    :return: Statistics on the occurrence of different mutual arrangement of tasks
    """

    fs12, fs21, ss12, ss21 = 0, 0, 0, 0
    ss12_lags, ss12_percent_lags, ss21_lags, ss21_percent_lags = [], [], [], []

    ffs12, ffs21 = 0, 0
    ffs12_lags, ffs12_percent_lags, ffs21_lags, ffs21_percent_lags = [], [], [], []

    if s1 == s2 and f1 == f2:
        ffs12 += 1
        ffs12_percent_lags.append(0.01)
        ffs12_lags.append(0.01)
    if f2 <= s1:
        fs21 += 1
    else:
        if s2 >= f1:
            fs12 += 1
        else:
            if s2 >= s1:
                if f2 >= f1:
                    ffs12 += 1
                    if get_delta_between_dates(f1, s1) != 0:
                        ffs12_percent_lags.append(get_delta_between_dates(s2, s1) / get_delta_between_dates(f1, s1))
                    else:
                        ffs12_percent_lags.append(0)
                    ffs12_lags.append(get_delta_between_dates(s2, s1))
                else:
                    ss12 += 1
                    if get_delta_between_dates(f1, s1) != 0:
                        ss12_percent_lags.append(get_delta_between_dates(s2, s1) / get_delta_between_dates(f1, s1))
                    else:
                        ss12_percent_lags.append(0)
                    ss12_lags.append(get_delta_between_dates(s2, s1))
            else:
                if f2 <= f1:
                    ffs21 += 1
                    if get_delta_between_dates(f2, s2) != 0:
                        ffs21_percent_lags.append(get_delta_between_dates(s1, s2) / get_delta_between_dates(f2, s2))
                    else:
                        ffs21_percent_lags.append(0)
                    ffs21_lags.append(get_delta_between_dates(s1, s2))
                else:
                    ss21 += 1
                    if get_delta_between_dates(f2, s2) != 0:
                        ss21_percent_lags.append(get_delta_between_dates(s1, s2) / get_delta_between_dates(f2, s2))
                    else:
                        ss21_percent_lags.append(0)
                    ss21_lags.append(get_delta_between_dates(s1, s2))
    return fs12, fs21, ss12, ss12_lags, ss12_percent_lags, ss21, ss21_lags, ss21_percent_lags, ffs12, ffs12_lags, ffs12_percent_lags, \
        ffs21, ffs21_lags, ffs21_percent_lags


def get_all_seq_statistic(history_data: pd.DataFrame, tasks_structure_df: pd.DataFrame, use_model_name: bool = False,
                          mapper: NameMapper = None) -> dict[int, list]:
    df_grouped = history_data.copy()

    if use_model_name:
        column_name = 'model_name'
    else:
        column_name = 'granular_smr_name'

    df_grouped = df_grouped.groupby('upper_works')[column_name].apply(list).reset_index(name="Works")
    works1, works2 = get_all_connections(tasks_structure_df, use_model_name, mapper)

    tasks_names = list(zip(works1['names'], works2['names']))
    tasks_ids = list(zip(works1['ids'], works2['ids']))

    predecessors_info_dict = {w_id: [] for w_id in tasks_structure_df['activity_id']}

    if len(tasks_names) != 0:
        for i in range(len(tasks_names)):
            w1, w2 = tasks_names[i]
            w1_id, w2_id = tasks_ids[i]

            if w1 != w2:
                fs12, fs21, ss12, ss21 = 0, 0, 0, 0
                ss12_lags, ss12_percent_lags, ss21_lags, ss21_percent_lags = [], [], [], []

                ffs12, ffs21 = 0, 0
                ffs12_lags, ffs12_percent_lags, ffs21_lags, ffs21_percent_lags = [], [], [], []

                for i, works_list in df_grouped.iterrows():
                    if w1 in works_list and w2 in works_list:
                        ind1 = history_data.loc[(history_data['upper_works'] == works_list['upper_works']) & (
                                    history_data['granular_smr_name'] == w1)]
                        ind2 = history_data.loc[(history_data['upper_works'] == works_list['upper_works']) & (
                                    history_data['granular_smr_name'] == w2)]

                        ind1_sorted = ind1.sort_values(by=['first_day', 'last_day']).reset_index(drop=True)
                        ind2_sorted = ind2.sort_values(by=['first_day', 'last_day']).reset_index(drop=True)

                        for l in range(min(len(ind1_sorted), len(ind2_sorted))):
                            s1, f1 = ind1_sorted.loc[l, 'first_day'], ind1_sorted.loc[l, 'last_day']

                            s2, f2 = ind2_sorted.loc[l, 'first_day'], ind2_sorted.loc[l, 'last_day']

                            if not any([type(x) == float for x in [s1, s2, f1, f2]]):
                                tasks_fs12, tasks_fs21, tasks_ss12, tasks_ss12_lags, tasks_ss12_percent_lags, tasks_ss21, tasks_ss21_lags, \
                                    tasks_ss21_percent_lags, tasks_ffs12, tasks_ffs12_lags, tasks_ffs12_percent_lags, tasks_ffs21, tasks_ffs21_lags, tasks_ffs21_percent_lags = gather_links_types_statistics(
                                    s1, f1, s2, f2)

                                fs12 += tasks_fs12
                                fs21 += tasks_fs21

                                ss12 += tasks_ss12
                                ss12_lags.extend(tasks_ss12_lags)
                                ss12_percent_lags.extend(tasks_ss12_percent_lags)
                                ss21 += tasks_ss21
                                ss21_lags.extend(tasks_ss21_lags)
                                ss21_percent_lags.extend(tasks_ss21_percent_lags)

                                ffs12 += tasks_ffs12
                                ffs12_lags.extend(tasks_ffs12_lags)
                                ffs12_percent_lags.extend(tasks_ffs12_percent_lags)
                                ffs21 += tasks_ffs21
                                ffs21_lags.extend(tasks_ffs21_lags)
                                ffs21_percent_lags.extend(tasks_ffs21_percent_lags)

                # Checking the mutual arrangement of tasks (follower - predecessor). We calculate which interposition is observed more often, correct the connections
                if fs12 + ffs12 + ss12 >= fs21 + ffs21 + ss21:
                    order_con = 1
                    fs = fs12
                    ffs = ffs12
                    ss = ss12
                else:
                    order_con = 2
                    fs = fs21
                    ffs = ffs21
                    ss = ss21

                if max([fs, ss, ffs]) != 0:
                    if fs > ss:
                        if ffs > 0:
                            if order_con == 1:
                                predecessors_info_dict[w2_id].append((w1_id, 'FFS',
                                                                      find_min_without_outliers(ffs12_percent_lags)))
                            else:
                                predecessors_info_dict[w1_id].append((w2_id, 'FFS',
                                                                      find_min_without_outliers(ffs21_percent_lags)))
                        else:
                            if order_con == 1:
                                predecessors_info_dict[w2_id].append((w1_id, 'FS', -1))
                            else:
                                predecessors_info_dict[w1_id].append((w2_id, 'FS', -1))
                    elif ss > ffs:
                        if order_con == 1:
                            predecessors_info_dict[w2_id].append((w1_id, 'SS',
                                                                  find_min_without_outliers(ss12_percent_lags)))
                        else:
                            predecessors_info_dict[w1_id].append((w2_id, 'SS',
                                                                  find_min_without_outliers(ss21_percent_lags)))
                    else:
                        if order_con == 1:
                            predecessors_info_dict[w2_id].append((w1_id, 'FFS',
                                                                  find_min_without_outliers(ffs12_percent_lags)))
                        else:
                            predecessors_info_dict[w1_id].append((w2_id, 'FFS',
                                                                  find_min_without_outliers(ffs21_percent_lags)))

    return predecessors_info_dict


def set_connections_info(structure_df: pd.DataFrame, history_data: pd.DataFrame, use_model_name: bool = False, mapper=None) \
        -> pd.DataFrame:
    """
    Restore tasks' connection based on history data

    :return: repaired DataFrame
    """
    tasks_df = structure_df.copy().set_index('activity_id')
    connections_dict = get_all_seq_statistic(history_data, structure_df, use_model_name, mapper)

    predecessors_ids_lst, predecessors_types_lst, predecessors_lags_lst = [], [], []

    for task_id, pred_info_lst in connections_dict.items():
        pred_ids_lst, pred_types_lst, pred_lags_lst = zip(*pred_info_lst)

        predecessors_ids_lst.append(','.join(map(str, pred_ids_lst)))
        predecessors_types_lst.append(','.join(map(str, pred_types_lst)))
        predecessors_lags_lst.append(','.join(map(str, pred_lags_lst)))

    # Convert strings to arrays
    tasks_df['predecessor_ids'] = predecessors_ids_lst
    tasks_df['connection_types'] = predecessors_types_lst
    tasks_df['lags'] = predecessors_lags_lst

    return tasks_df
